import os
import sys
sys.path.insert(0, os.path.abspath('..'))

project = 'LabPaper'
copyright = '2024-2025'
author = 'Khris Griffis, Ph.D.'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'myst_parser',
]

# Configure MyST-Parser to enable all extensions
myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The theme to use for HTML and HTML Help pages.
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# Intersphinx configuration
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'nbconvert': ('https://nbconvert.readthedocs.io/en/latest/', None),
    'jupyter': ('https://jupyter.readthedocs.io/en/latest/', None),
}

# AutoDoc configuration
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__'
} 


import re

def parse_metadata_content(source: str, is_markdown: bool = False) -> tuple[dict, str]:
    """Parse metadata content with continuation lines."""
    metadata = {}
    updated_source = source.splitlines()
    to_remove = []

    # Define patterns based on cell type
    allowed_tags = "|".join(["caption", "label", "footer", "width"])
    footnote_pattern = r'(?:\[(\d+(?:,\d+)*)\]\s*)?(.+?\|?)$'
    if is_markdown:
        single_pattern = r'<!--\s*@(%s):\s*([^|]+?)\s*-->$'
        multi_start = r'<!--\s*(?:\n)?\s*@(%s):\s*(.+?\s*\|)\s*-->$'
        multi_subsequent = r'<!--\s*(?:\[(\d+(?:,\d+)*)\]\s*)?(.+?)\s*-->$'
    else:
        single_pattern = r'#\s*@(%s):\s*([^|]+?)\s*$'
        multi_start = r'#\s*@(%s):\s*(.+?)\s*\|$'
        multi_subsequent = r'#\s*(?:\[(\d+(?:,\d+)*)\]\s*)?(.+?)$'
    # Initialize variables
    current_type = None
    current_content = []
    in_multiline = False
    
    for i, line in enumerate(updated_source):
        stripped = line.strip()
        if not stripped:
            # Empty lines indicate either end of tag or end of content
            # Store the contents if any, otherwise move on
            if current_type:
                if current_type == "footer":
                    metadata[current_type] = current_content
                else:
                    metadata[current_type] = "\n".join(current_content)
                current_type = None
                current_content = []
            continue
        
        if not in_multiline:
            # check for single line matches
            match = re.match(single_pattern % allowed_tags, stripped)
            if match:
                # Found single pattern with without line break ending
                key, value = match.group(1).strip(), match.group(2).strip()
                metadata[key] = value
                to_remove.append(i)
                continue
            # Check for multi-line opening match
            match = re.match(multi_start % allowed_tags, stripped)
            if match:
                # Found multi-line opening match
                current_type, value = match.group(1).strip(), match.group(2).strip().rstrip("|")
                if current_type == "footer":
                    submatch = re.match(footnote_pattern, value)
                    footnote = {'text':submatch.group(2).strip()}
                    if submatch.group(1):
                        footnote['number'] = submatch.group(1).strip()
                    current_content.append(footnote)
                else:
                    current_content = [value]
                # Set the multiline flag
                in_multiline = True
                to_remove.append(i)
                continue
        else:
            # In multiline mode,
            match = re.match(multi_subsequent, stripped)
            if match:
                num, text = match.group(1), match.group(2).strip()
                if current_type == "footer":
                    footnote = {"text": text.rstrip("|")}
                    if num:
                        footnote["number"] = num
                    current_content.append(footnote)
                else:
                    current_content.append(text.rstrip("|"))
                to_remove.append(i)
                # End of multiline if no `|`
                if not text.endswith("|"):
                    in_multiline = False
                    if current_type == "footer":
                        metadata[current_type] = current_content
                    else:
                        metadata[current_type] = "\n".join(current_content)
                    current_type = None
                    current_content = []
            else:
                print("No match found")
    # Remove processed lines in reverse order
    for i in sorted(to_remove, reverse=True):
        del updated_source[i]
    
    return metadata, '\n'.join(updated_source)


"""LabPaper - A Jupyter notebook exporter for academic papers."""

__version__ = "1.0.12"

from .exporters import SpringerNaturePDF
from .preprocessors import PythonMarkdownPreprocessor, PygmentizePreprocessor

__all__ = [
    "SpringerNaturePDF",
    "PythonMarkdownPreprocessor",
    "PygmentizePreprocessor"
] 


class LatexFailed(IOError):
    """Exception for failed latex run.
    
    From nbconvert.exporters.pdf.LatexFailed
    """
    def __init__(self, output):
        self.output = output

    def __str__(self):
        return f'PDF creating failed, captured latex output:\n{self.output}'


"""
Preprocessors for handling different input formats.
"""
from traitlets import Unicode, List

from nbconvert.preprocessors import Preprocessor
from nbconvert.preprocessors.execute import CellExecutionError
from pygments.formatters import LatexFormatter

import re
from contextlib import redirect_stdout, redirect_stderr
import io
import logging

import matplotlib.pyplot as plt  # Import pyplot here
import matplotlib
matplotlib.use('Agg')  # Set non-interactive backend before any other matplotlib imports

class PygmentizePreprocessor(Preprocessor):
    
    style = Unicode("abap", help="Name of the pygments style to use").tag(config=True)
    
    def preprocess(self, nb, resources):
        # Generate Pygments definitions for Latex
        resources.setdefault("latex", {})
        resources["latex"].setdefault(
            "pygments_definitions", LatexFormatter(style=self.style).get_style_defs()
        )
        resources["latex"].setdefault("pygments_style_name", self.style)
        return nb, resources

class PythonMarkdownPreprocessor(Preprocessor):
    def __init__(self, **kw):
        super().__init__(**kw)
        self._global_ns = {'plt': plt}  # Initialize global_ns here

    @property
    def global_ns(self):
        return self._global_ns

    date = Unicode(
        None,
        help=("Date of the LaTeX document"),
        allow_none=True,
    ).tag(config=True)

    title = Unicode(None, help=("Title of the LaTeX document"), allow_none=True).tag(config=True)

    author_names = List(
        Unicode(),
        default_value=None,
        help=("Author names to list in the LaTeX document"),
        allow_none=True,
    ).tag(config=True)

    def has_inline(self, cell):
        if cell.cell_type == 'markdown':
            return bool(re.search(r'\{\{\s*(.*?)\s*\}\}', cell.source))
        elif cell.cell_type == 'code':
            return bool(re.search(r'#.*\{\{\s*(.*?)\s*\}\}.*', cell.source))
        return False

    def execute_code(self, code, cell_index):
        try:
            stdout_buffer = io.StringIO()
            stderr_buffer = io.StringIO()
            with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
                exec(code, self.global_ns)
                plt.close('all')
            stdout_content = stdout_buffer.getvalue()
            stderr_content = stderr_buffer.getvalue()
            if stderr_content:
                self.log.critical(f"Cell {cell_index} stderr:\n{stderr_content}")
            return stdout_content
        except CellExecutionError as e:
            self.log.warning(f"Execution error in cell {cell_index}: {str(e)}")
        except Exception as e:
            self.log.warning(f"Failed to execute cell {cell_index}: {str(e)}")
        return ""

    def execute_cell(self, cell, cell_index):
        if cell.cell_type == 'code' and ('skip-execution' not in cell.metadata.get('tags', [])):
            self.execute_code(cell.source, cell_index)

    def evaluate_expression(self, expr, cell_index):
        try:
            stdout_buffer = io.StringIO()
            stderr_buffer = io.StringIO()
            with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
                result = eval(expr, self.global_ns)
            return str(result)
        except Exception as e:
            self.log.warning(f"Warning: Failed to evaluate expression '{expr}' in cell {cell_index}: {str(e)}")
            return f"{{{{ {expr} }}}}"

    def execute_inline(self, cell, cell_index):
        pattern = r'\{\{\s*(.*?)\s*\}\}'

        def evaluate_expression(match):
            expr = match.group(1)
            return self.evaluate_expression(expr, cell_index)

        if cell.cell_type == 'markdown':
            cell.source = re.sub(pattern, evaluate_expression, cell.source)
        elif cell.cell_type == 'code':
            lines = cell.source.split('\n')
            processed_lines = []
            for line in lines:
                stripped = line.lstrip()
                if stripped.startswith('#'):
                    comment_text = stripped[1:]
                    processed_comment = re.sub(pattern, evaluate_expression, comment_text)
                    processed_line = line[:len(line)-len(stripped)] + '#' + processed_comment
                    processed_lines.append(processed_line)
                else:
                    processed_lines.append(line)
            cell.source = '\n'.join(processed_lines)

    def preprocess(self, nb, resources):
        if self.author_names is not None:
            nb.metadata["authors"] = [{"name": author} for author in self.author_names]
        if self.date is not None:
            nb.metadata["date"] = self.date
        if self.title is not None:
            nb.metadata["title"] = self.title

        # First pass: check for inline expressions
        has_expr = False
        for cell in nb.cells:
            has_expr = self.has_inline(cell)
            if has_expr:
                break
        if not has_expr:
            return nb, resources

        # Second pass: execute code cells to build up the namespace
        for index, cell in enumerate(nb.cells):
            self.execute_cell(cell, index)

        # Third pass: process inline contents
        for index, cell in enumerate(nb.cells):
            self.execute_inline(cell, index)

        return nb, resources



"""Custom filters to be used with IPython nbconvert."""
import re

__all__ = ['newline_block', 'latex_internal_references','comment_lines']

def newline_block(text):
    """ Filter that makes sure that a block will always start with a newline """
    return r"""
{0:s}
""".format(text)


def latex_internal_references(text):
    """Take markdown text and replace instances of
    '[blah](#anchor-ref)' with 'autoref{anchor-ref}'
    """
    ref_pattern = re.compile(r'\[(?P<text>.*?)\]\(#(?P<reference>.*?)\)')
    replacement = r'\\autoref{\g<reference>}'

    return ref_pattern.sub(replacement, text)

def comment_lines(text, comment_char="%"):
    """Adds a comment character and space to every line of the text."""
    lines = text.splitlines()
    return "\n".join(f"{comment_char} {line}" for line in lines)


"""Address handling utilities."""
from dataclasses import dataclass
import usaddress


@dataclass
class Address:
    department: str = ''
    organization: str = ''
    street: str = ''
    city: str = ''
    state: str = ''
    postcode: str = ''
    country: str = ''
    
    @classmethod
    def from_string(cls, address_str: str):
        """Parse address from string using usaddress."""
        parts = [p.strip() for p in address_str.split(',')]
        addr = cls()
        
        # First part is always department
        if parts:
            addr.department = parts[0]
        
        # Second part is organization
        if len(parts) > 1:
            addr.organization = parts[1]
        
        # Try to parse the remaining parts as an address
        if len(parts) > 2:
            try:
                address_string = ', '.join(parts[2:])
                tagged_address, _ = usaddress.tag(address_string)
                
                addr.street = tagged_address.get('AddressNumber', '') + ' ' + \
                            tagged_address.get('StreetName', '') + ' ' + \
                            tagged_address.get('StreetNamePostType', '')
                addr.street = addr.street.strip()
                
                addr.city = tagged_address.get('PlaceName', '')
                addr.state = tagged_address.get('StateName', '')
                addr.postcode = tagged_address.get('ZipCode', '')
                
                # Country is usually the last part
                if len(parts) > 3:
                    addr.country = parts[-1].strip()
                
            except usaddress.RepeatedLabelError:
                # Fallback to simple parsing if address is complex
                if len(parts) > 2:
                    addr.street = parts[2]
                if len(parts) > 3:
                    addr.city = parts[3]
                if len(parts) > 4:
                    addr.state = parts[4]
                if len(parts) > 5:
                    addr.postcode = parts[5]
                if len(parts) > 6:
                    addr.country = parts[6]
        
        return addr
    
    def __eq__(self, other):
        """Compare two addresses for equality."""
        if not isinstance(other, Address):
            return NotImplemented
        return (
            self.department == other.department and
            self.organization == other.organization and
            self.street == other.street and
            self.city == other.city and
            self.state == other.state and
            self.postcode == other.postcode and
            self.country == other.country
        )

    def __hash__(self):
        """Hash function for Address instances."""
        return hash((
            self.department,
            self.organization,
            self.street,
            self.city,
            self.state,
            self.postcode,
            self.country
        ))

    def __repr__(self):
        """Detailed string representation."""
        parts = []
        if self.department:
            parts.append(f"dept='{self.department}'")
        if self.organization:
            parts.append(f"org='{self.organization}'")
        if self.street:
            parts.append(f"street='{self.street}'")
        if self.city:
            parts.append(f"city='{self.city}'")
        if self.state:
            parts.append(f"state='{self.state}'")
        if self.postcode:
            parts.append(f"postcode='{self.postcode}'")
        if self.country:
            parts.append(f"country='{self.country}'")
        
        return f"Address({', '.join(parts)})"


"""Exporters for LabPaper."""
from .base import LabPaperBaseExporter
from .springer import SpringerNaturePDF
from .latexerror import LatexFailed

__all__ = ['LabPaperBaseExporter', 'SpringerNaturePDF', 'LatexFailed'] 


"""Springer Nature PDF exporter."""
import os
import re
import copy
from traitlets import Unicode, List, Bool, default
from traitlets.config import Config
from nbformat.notebooknode import NotebookNode

from .base import LabPaperBaseExporter
from .utils.author import parse_authors_section
from .utils.code_parser import parse_metadata_content
from .utils.resolvers import resolve_boolean, resolve_string

# TODO: Add support for markdown images as figures

class SpringerNaturePDF(LabPaperBaseExporter):
    """Springer Nature PDF exporter that uses sn-jnl class."""
    
    _figure_count = 0
    _table_count = 0
    export_from_notebook = "Springer Nature PDF"
    output_mimetype = "application/pdf"
    
    # Metadata processing configuration
    process_metadata = Bool(
        True,
        help="Whether to process and transform cell metadata"
    ).tag(config=True)
    
    execute_notebook = Bool(
        True,
        help="Whether to execute the notebook before processing"
    ).tag(config=True)
    
    bibliography_style = Unicode(
        default_value='sn-nature',
        help="BibTeX style to use for citations"
    ).tag(config=True)
    
    latex_class = Unicode(
        default_value='sn-jnl',
        help="LaTeX class to use for the document"
    ).tag(config=True)
    
    latex_engine = Unicode(
        default_value='pdflatex',
        help="LaTeX engine to use (pdflatex recommended)",
    ).tag(config=True)
    
    latex_command = List(
        trait=Unicode(),
        default_value=["{filename}", "-shell-escape", "-interaction=nonstopmode"],
        help="Shell command used to compile latex."
    ).tag(config=True)

    bib_command = List(
        ["bibtex", "{filename}"],
        help="Shell command used to run bibtex."
    ).tag(config=True)
    
    figure_placement = Unicode(
        default_value='htbp',
        help="LaTeX placement for figures"
    ).tag(config=True)
    
    separate_fig_and_tab_page = Bool(
        default_value=False,
        help="Whether to separate figures and tables onto separate pages using the endfloat package"
    ).tag(config=True)
    
    referee = Bool(
        default_value=False,
        help="Whether to use the referee option for the document class"
    ).tag(config=True)
    
    line_numbers = Bool(
        default_value=False,
        help="Whether to show line numbers in the code blocks"
    ).tag(config=True)
    
    no_bib = Bool(
        default_value=False,
        help="Whether to suppress the bibliography"
    ).tag(config=True)
    
    @property
    def default_config(self):
        """Default configuration for LabPaper base exporter"""
        c = Config({
            'labpaper.preprocessors.PygmentizePreprocessor': {'enabled': True},
            'ExecutePreprocessor': {'enabled': self.execute_notebook,'metadata':{'path':"./"}}
        })
        if super().default_config:
            c2 = super().default_config.copy()
            c2.merge(c)
            c = c2
        return c
    
    @default('metadata_tags')
    def _metadata_tags(self):
        return {
            'figure': ['figure', 'plot'],
            'table': ['table', 'tabular'],
            'stable': ['stable','sideways', 'sidewaystable'],
            'equation': ['equation', 'math'],
            'code': ['code', 'listing'],
            'hide': ['hide', 'hidden', 'no-display'],
            'show': ['show', 'display', 'visible', 'echo'],
        }
    
    @default('extra_template_basedirs')
    def _extra_template_basedirs(self):
        """Return the paths to the templates."""
        return [
            os.path.abspath(
                os.path.join(
                    os.path.dirname(
                        os.path.dirname(__file__)
                        ), 
                    "templates"
                    )
                )
        ]
    
    @default("file_extension")
    def _file_extension_default(self):
        return ".tex"
    
    @default("template_name")
    def _template_name_default(self):
        """Default to the nature template"""
        return 'nature'

    @default('template_file')
    def _template_file_default(self):
        """Default to nature/index.tex.j2"""
        return 'index.tex.j2'
    
    def __init__(self, config=None, **kw):
        default_config = self.default_config
        if config:
            default_config.merge(config)
        
        config_traits = self._resolve_traits(default_config)
        super().__init__(config=config_traits, **kw)
   
    def _resolve_traits(self, config):
        """
        Resolve traitlets configurations.
        Property -> Configs conversions
        """
        # Resolve execute_notebook to a boolean
        config.ExecutePreprocessor.enabled = resolve_boolean(
            config.SpringerNaturePDF.execute_notebook,
            self.execute_notebook
            )
        # Resolve figure placement to a string
        self.figure_placement = resolve_string(
            config.SpringerNaturePDF.figure_placement,
            self.figure_placement
        )
        # Resolve separate_fig_and_tab_page to a boolean
        self.separate_fig_and_tab_page = resolve_boolean(
            config.SpringerNaturePDF.separate_fig_and_tab_page,
            self.separate_fig_and_tab_page
        )
        # Resolve referee to a boolean
        self.referee = resolve_boolean(
            config.SpringerNaturePDF.referee,
            self.referee
        )
        # Resolve line_numbers to a boolean
        self.line_numbers = resolve_boolean(
            config.SpringerNaturePDF.line_numbers,
            self.line_numbers
        )
        # Resolve bibliography_style to a string
        bibstyle = resolve_string(
            config.SpringerNaturePDF.bibliography_style,
            self.bibliography_style
        )
        # Resolve no_bib to a boolean
        self.no_bib = resolve_boolean(
            config.SpringerNaturePDF.no_bib,
            self.no_bib
        )
        # Dynamically parse .bst file names from texmf/bibtex/bst/ using self._get_texmf_path()
        texmf_bst_directory = os.path.join(self._get_texmf_path(), 'bibtex', 'bst')
        valid_bibstyles = []
        for root, _, files in os.walk(texmf_bst_directory):
            for file in files:
                if file.endswith('.bst'):
                    valid_bibstyles.append(os.path.splitext(file)[0])

        if bibstyle not in valid_bibstyles:
            self.log.warning(f"Invalid bibliography style '{bibstyle}', defaulting to 'sn-nature'")
            bibstyle = 'sn-nature'
        self.bibliography_style = bibstyle
        
        return config
    
    # Parent Overrides
    def process_cell_tags(self, tags):
        """Process cell tags according to Springer Nature mappings."""
        processed = {category: False for category in self._metadata_tags()}
        
        # Set hide to True by default
        processed['hide'] = True
        
        for tag in tags:
            # Process other tags
            for category, matches in self._metadata_tags().items():
                if tag in matches:
                    processed[category] = True
                    break
            
            # Check for show tags first
            if tag in self._metadata_tags()['show']:
                processed['hide'] = False
                continue
            
        
        return processed
        
    # Base methods for subclasses to implement, these tend to need super() calls
    def process_notebook_metadata(self, nb):
        """Process notebook-level metadata."""
        nb = super().process_notebook_metadata(nb)
        metadata = self._get_metadata_handle(nb) # labpaper
        
        # Check first cell for front matter
        if len(nb.cells) > 0 and nb.cells[0].cell_type == 'markdown':
            first_cell = nb.cells[0]
            
            # Look for title in first level header
            title_match = re.match(r'^#\s+(.+)$', first_cell.source, re.MULTILINE)
            if title_match:
                title = title_match.group(1).strip()
                self.log.info(f"Found title in first cell: {title}")
                metadata.title = title
                # Remove the title line from the cell source
                first_cell.source = re.sub(r'^#\s+.+\n?', '', first_cell.source, flags=re.MULTILINE)
            
            # Parse authors and affiliations
            authors, affiliations = parse_authors_section(first_cell.source)
            if authors:
                metadata.authors = [NotebookNode(author.__dict__) for author in authors]
                metadata.affiliations = [NotebookNode(affil.__dict__) for affil in affiliations]
                
            # Parse other metadata (abstract, keywords, date)
            abstract_match = re.search(
                r'^Abstract:\s*\n((?:.|\n)*?)(?=\n^(Keywords:|Date:)|\Z)',
                first_cell.source,
                re.MULTILINE
            )
            if abstract_match:
                metadata.abstract = abstract_match.group(1).strip()
                
            keywords_match = re.search(
                r'^Keywords:\s*(.+?)(?:\n\n|$)',
                first_cell.source,
                re.MULTILINE
            )
            if keywords_match:
                metadata.keywords = [k.strip() for k in keywords_match.group(1).split(',')]
                
            date_match = re.search(
                r'^Date:\s*(.+?)(?:\n\n|$)',
                first_cell.source,
                re.MULTILINE
            )
            if date_match:
                metadata.date = date_match.group(1).strip()
            # Remove the first cell contents after parsing
            first_cell.source = ''

        # Process remaining front matter
        front_matter = nb.metadata.get('front_matter', {})
        metadata.update({
            'title': metadata.get('title', front_matter.get('title', '')),
            'author': (f"{metadata.authors[0].given_name} {metadata.authors[0].surname}" 
                      if metadata.get('authors') else front_matter.get('author', '')),
            'date': metadata.get('date', front_matter.get('date', '\\today')),
            'abstract': metadata.get('abstract', front_matter.get('abstract', '')),
            'keywords': metadata.get('keywords', front_matter.get('keywords', '')),
            'document_class': 'sn-jnl',
            'separate_fig_and_tab_page': self.separate_fig_and_tab_page,
            'figure_placement': self.figure_placement,
            'referee': self.referee,
            'line_numbers': self.line_numbers,
            'bibliography_style': self.bibliography_style
        })
        
        
        return nb
    
    def process_cell_metadata(self, nb): 
        """Process metadata for all cells."""
        nb = super().process_cell_metadata(nb)
        
        # Loop through cells, gather processed tags, parse figures and tables
        for cell in nb.cells:
            metadata = self._get_metadata_handle(cell)
            tags = metadata.get('processed_tags', {})
            
            # Process figures
            if self._has_figure_content(cell) or tags.get('figure', False):
                # Process code
                if tags.get('show', False):
                    self._process_code(cell, metadata, 'fig')
                else:
                    self._process_figure(cell, metadata)
            
            # Process tables
            if tags.get('table', False) or tags.get('stable', False):
                # Process code
                if tags.get('show', False):
                    self._process_code(cell, metadata, 'tab')
                else:
                    self._process_table(cell, metadata)
            
            # Process code
            if tags.get('show', False) and not (tags.get('figure', False) or tags.get('table', False) or tags.get('stable', False)):
                self._process_code(cell, metadata, 'code')
                
        return nb
    
    # FIGURE PROCESSING -------------------------------------------------------#
    def _has_figure_content(self, cell):
        if cell.cell_type == 'markdown':
            # Check for markdown image syntax with common image extensions
            return bool(re.search(
                r'!\[.*?\]\(.*?\.(?:png|jpe?g|gif|tiff?|bmp|eps|pdf)\)',
                cell.source,
                re.IGNORECASE
            ))
        elif cell.cell_type == 'code' and hasattr(cell, 'outputs'):
            return any(
                output.get('output_type') == 'display_data' and 
                any(mime.startswith('image/') for mime in output.get('data', {}))
                for output in cell.outputs
            )
        return False
    
    def _process_figure(self, cell, metadata):
        """Process figure metadata."""
        float_metadata, updated_source = parse_metadata_content(cell.source, is_markdown=cell.cell_type == 'markdown')
        cell.source = updated_source
        
        self._figure_count += 1
        
        # Generate label if not provided
        if not float_metadata.get('label'):
            float_metadata['label'] = f'fig:{self._figure_count:02d}'
        
        if cell.cell_type == 'markdown':
            metadata.update(float_metadata)
        else:
            # Process code cell outputs
            for output in cell.outputs:
                if 'metadata' not in output:
                    output.metadata = NotebookNode()
                output.metadata.update(float_metadata)
        return copy.deepcopy(float_metadata)
                    
    
    # TABLE PROCESSING --------------------------------------------------------#
    def _process_table(self, cell, cell_metadata):
        """Process table metadata."""
        float_metadata, updated_source = parse_metadata_content(cell.source, is_markdown=cell.cell_type == 'markdown')
        cell.source = updated_source
        
        self._table_count += 1
        
        # Generate label if not provided
        if not float_metadata.get('label'):
            float_metadata['label'] = f'tab:{self._table_count}'
        
        if cell.cell_type == 'markdown':
            cell_metadata.update(float_metadata)
        else:
            # Process code cell outputs
            for output in cell.outputs:
                if 'metadata' not in output:
                    output.metadata = NotebookNode()
                output.metadata.update(float_metadata)
        return copy.deepcopy(float_metadata)
    
    # CODE PROCESSING --------------------------------------------------------#
    def _process_code(self, cell, cell_metadata, float_type):
        """Process code metadata."""
        # Increment figure counter for all code blocks
        self._figure_count += 1
        current_float_count = self._figure_count

        # Handle figures and tables
        if float_type == 'fig':
            float_metadata = self._process_figure(cell, cell_metadata)
            float_metadata['caption'] = f'Source code for Fig.~\\ref{{{float_metadata["label"]}}}'
        elif float_type == 'tab':
            float_metadata = self._process_table(cell, cell_metadata)
            float_metadata['caption'] = f'Source code for Table~\\ref{{{float_metadata["label"]}}}'
        else:
            # Standalone code: parse metadata and update source
            float_metadata, updated_source = parse_metadata_content(cell.source, is_markdown=False)
            cell.source = updated_source
            float_metadata.setdefault('label', f'code:{current_float_count:02d}')
            float_metadata.setdefault('caption', 'Source Code.')

        # Update metadata
        if float_type in ('fig', 'tab'):
            float_metadata['label'] = f'code:{float_metadata["label"]}'

        cell_metadata.update(float_metadata)
        return copy.deepcopy(float_metadata)
